"""Test helpers."""
from unittest.mock import AsyncMock

import pytest

from mqtt_entity import Device, SensorEntity
from mqtt_entity.helpers import hass_default_rw_icon, hass_device_class, set_attributes


def test_helpers():
    """Test helpers."""
    assert hass_device_class(unit="kWh") == "energy"
    assert hass_default_rw_icon(unit="W") == "mdi:flash"


@pytest.mark.asyncio
async def test_set_attributes():
    """Test set_attributes."""
    dev = Device(identifiers=["test123"])
    e = SensorEntity(
        attributes_topic="blah",
        unique_id="a1",
        device=dev,
        state_topic="/st",
        name="test1",
    )
    mc = AsyncMock()
    thea = {"the": "attr"}
    await set_attributes(thea, entity=e, client=mc)
    assert mc.publish.call_count == 1
    assert mc.publish.call_args[1]["topic"] == "blah"



"""MQTT entities."""
import inspect
import logging
from typing import Any, Callable, Optional, Sequence, Union

import attrs
from attrs import validators

from mqtt_entity.utils import BOOL_OFF, BOOL_ON, required

_LOGGER = logging.getLogger(__name__)

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@attrs.define
class Device:
    """A Home Assistant Device, used to group entities."""

    identifiers: list[Union[str, tuple[str, Any]]] = attrs.field(
        validator=[validators.instance_of(list), validators.min_len(1)]
    )
    connections: list[str] = attrs.field(factory=list)
    configuration_url: str = attrs.field(default="")
    manufacturer: str = attrs.field(default="")
    model: str = attrs.field(default="")
    name: str = attrs.field(default="")
    suggested_area: str = attrs.field(default="")
    sw_version: str = attrs.field(default="")
    via_device: str = attrs.field(default="")

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """The device identifier."""
        return str(self.identifiers[0])


@attrs.define
class Availability:
    """Represent Home Assistant entity availability."""

    topic: str = attrs.field()
    payload_available: str = attrs.field(default="online")
    payload_not_available: str = attrs.field(default="offline")
    value_template: str = attrs.field(default="")


@attrs.define
class Entity:
    """A generic Home Assistant entity used as the base class for other entities."""

    unique_id: str = attrs.field()
    device: Device = attrs.field()
    state_topic: str = attrs.field()
    name: str = attrs.field()
    availability: list[Availability] = attrs.field(factory=list)
    availability_mode: str = attrs.field(default="")
    device_class: str = attrs.field(default="")
    unit_of_measurement: str = attrs.field(default="")
    state_class: str = attrs.field(default="")
    expire_after: int = attrs.field(default=0)
    """Unavailable if not updated."""
    enabled_by_default: bool = attrs.field(default=True)
    entity_category: str = attrs.field(default="")
    icon: str = attrs.field(default="")
    attributes_topic: str = attrs.field(default="")
    """Used by the set_attributes helper."""

    _path = ""

    def __attrs_post_init__(self) -> None:
        """Init the class."""
        if not self._path:
            raise TypeError(f"Do not instantiate {self.__class__.__name__} directly")
        if not self.state_class and self.device_class == "energy":
            self.state_class = "total_increasing"

    @property
    def asdict(self) -> dict:
        """Represent the entity as a dictionary, without empty values and defaults."""

        def _filter(atrb: attrs.Attribute, value: Any) -> bool:
            return (
                bool(value) and atrb.default != value and not inspect.isfunction(value)
            )

        return attrs.asdict(self, filter=_filter)

    @property
    def topic(self) -> str:
        """Discovery topic."""
        uid, did = self.unique_id, self.device.id
        if uid.startswith(did):
            uid = uid[len(did) :].strip("_")
        return f"homeassistant/{self._path}/{did}/{uid}/config"


@attrs.define
class SensorEntity(Entity):
    """A Home Assistant Sensor entity."""

    _path = "sensor"


@attrs.define
class BinarySensorEntity(Entity):
    """A Home Assistant Binary Sensor entity."""

    payload_on: str = attrs.field(default=BOOL_ON)
    payload_off: str = attrs.field(default=BOOL_OFF)

    _path = "binary_sensor"


@attrs.define
class RWEntity(Entity):
    """Read/Write entity base class.

    This will default to a text entity.
    """

    command_topic: str = attrs.field(
        default="", validator=(validators.instance_of(str), validators.min_len(2))
    )

    on_change: Optional[Callable] = attrs.field(default=None)

    _path = "text"


@attrs.define
class SelectEntity(RWEntity):
    """A HomeAssistant Select entity."""

    options: Sequence[str] = attrs.field(default=None, validator=required)

    _path = "select"


@attrs.define
class SwitchEntity(RWEntity):
    """A Home Assistant Switch entity."""

    payload_on: str = attrs.field(default=BOOL_ON)
    payload_off: str = attrs.field(default=BOOL_OFF)

    _path = "switch"


@attrs.define
class NumberEntity(RWEntity):
    """A HomeAssistant Number entity."""

    min: float = attrs.field(default=0.0)
    max: float = attrs.field(default=100.0)
    mode: str = attrs.field(default="auto")
    step: float = attrs.field(default=1.0)

    _path = "number"



"""Test entities."""
import pytest

from mqtt_entity import (
    Availability,
    Device,
    Entity,
    NumberEntity,
    RWEntity,
    SensorEntity,
)


def test_ent():
    """Test entity."""
    with pytest.raises(TypeError) as err:
        Entity()
        assert "unique_id" in err
        assert "device" in err
        assert "state_topic" in err
        assert "name" in err
    kwa = {
        "unique_id": "1",
        "device": Device(identifiers=["d"]),
        "state_topic": "/top",
        "name": "a",
    }
    with pytest.raises(TypeError) as err:
        Entity(**kwa)
        assert " Entity directly" in err
    with pytest.raises(TypeError) as err:
        RWEntity(command_topic="/a", **kwa)
        assert " RWEntity directly" in err
    with pytest.raises(ValueError) as err:
        NumberEntity(**kwa)
        assert "command_topic" in err
        assert "mzzzissingx" in err
    NumberEntity(command_topic="/a", **kwa)


def test_dev():
    """Test device."""
    with pytest.raises(TypeError):
        Device()
    with pytest.raises(ValueError):
        Device(identifiers=[])
    Device(identifiers=["123"])


def test_mqtt_entity():
    """Test MQTT."""
    dev = Device(identifiers=["123"])

    ava = Availability(topic="/blah")

    ent = SensorEntity(
        name="test1",
        unique_id="789",
        device=dev,
        availability=[ava],
        state_topic="/test/a",
    )
    assert ent.asdict == {
        "name": "test1",
        "unique_id": "789",
        "device": {"identifiers": ["123"]},
        "availability": [{"topic": "/blah"}],
        "state_topic": "/test/a",
    }

    assert ent.topic == "homeassistant/sensor/123/789/config"



"""Helpers."""
from json import dumps
from typing import Any

from mqtt_entity.client import MQTTClient
from mqtt_entity.entities import Entity


def hass_default_rw_icon(*, unit: str) -> str:
    """Get the HASS default icon from the unit."""
    return {
        "W": "mdi:flash",
        "V": "mdi:sine-wave",
        "A": "mdi:current-ac",
        "%": "mdi:battery-lock",
    }.get(unit, "")


def hass_device_class(*, unit: str) -> str:
    """Get the HASS device_class from the unit."""
    return {
        "W": "power",
        "kW": "power",
        "kVA": "apparent_power",
        "VA": "apparent_power",
        "V": "voltage",
        "kWh": "energy",
        "kVAh": "",  # Not energy
        "A": "current",
        "°C": "temperature",
        "%": "battery",
    }.get(unit, "")


async def set_attributes(
    attributes: dict[str, Any],
    *,
    entity: Entity,
    client: MQTTClient,
    retain: bool = False,
) -> None:
    """Set attributes helper."""
    if not entity.attributes_topic:
        raise ValueError(f"Entity '{entity.name}' needs an attributes_topic.")
    await client.publish(
        topic=entity.attributes_topic,
        payload=dumps(attributes),
        retain=retain,
    )



"""Test MQTT class."""
import asyncio
import logging
from unittest.mock import MagicMock, Mock, patch

import pytest

import mqtt_entity
from mqtt_entity.client import _mqtt_on_connect

_LOGGER = logging.getLogger(__name__)

MQTT_HOST = "192.168.1.8"
MQTT_PASS = "hass123"
MQTT_USER = "hass"


@pytest.mark.asyncio
@pytest.mark.mqtt
async def test_mqtt_server():
    """Test MQTT."""
    mqc = mqtt_entity.MQTTClient()
    mqc.availability_topic = "test/available"
    await mqc.connect(username="hass", password="hass123", host="192.168.1.8")
    dev = mqtt_entity.Device(identifiers=["test123"])
    dev2 = mqtt_entity.Device(identifiers=["test789"])

    select_id = "t_select_1"
    select_id2 = "t_select_2"
    sense_id = "t_sense_1"

    async def select_select(msg):
        _LOGGER.error("onchange start: %s", msg)
        await mqc.publish(f"test/{select_id2}", "opt 4")
        await mqc.publish(f"test/{select_id}", msg)
        _LOGGER.error("onchange done: %s", msg)

    _loop = asyncio.get_running_loop()

    def select_select2(msg):
        _LOGGER.error("onchange no async: %s", msg)
        _loop.create_task(mqc.publish(f"test/{select_id2}", msg))
        _LOGGER.error("onchange no async done: %s", msg)

    ent = [
        mqtt_entity.SelectEntity(
            name="Test select entity",
            unique_id=select_id,
            device=dev,
            command_topic=f"test/{select_id}_set",
            options=["opt 1", "opt 2"],
            on_change=select_select,
            state_topic=f"test/{select_id}",
        ),
        mqtt_entity.SelectEntity(
            name="Test select entity 2",
            unique_id=select_id2,
            device=dev,
            command_topic=f"test/{select_id2}_set",
            options=["opt 3", "opt 4"],
            on_change=select_select2,
            state_topic=f"test/{select_id2}",
        ),
        mqtt_entity.SensorEntity(
            name="Test sensor entity",
            unique_id=sense_id,
            device=dev2,
            state_topic=f"test/{sense_id}",
        ),
    ]
    await mqc.publish_discovery_info(entities=ent)
    await asyncio.sleep(0.1)

    await mqc.publish(f"test/{select_id}", "opt 2")
    await mqc.publish(f"test/{select_id2}", "opt 3")
    await mqc.publish(f"test/{sense_id}", "yay!")
    for i in range(100):
        await asyncio.sleep(0.5)
    await mqc.remove_discovery_info(device_ids=[dev.id, dev2.id], keep_topics=[])

    await mqc.disconnect()
    await asyncio.sleep(0.1)

    assert False


@pytest.mark.asyncio
@pytest.mark.mqtt
async def test_mqtt_discovery():
    """Test MQTT."""
    root = "test2"
    mqc = mqtt_entity.MQTTClient()
    mqc.availability_topic = f"{root}/available"
    await mqc.connect(username=MQTT_USER, password=MQTT_PASS, host=MQTT_HOST)
    dev = mqtt_entity.Device(identifiers=[f"id_{root}"])

    sensor_id = [f"sen{i}" for i in range(3)]

    entities = [
        mqtt_entity.SensorEntity(
            name="Test select entity",
            unique_id=id,
            device=dev,
            state_topic=f"{root}/{id}",
        )
        for id in sensor_id
    ]

    await mqc.publish_discovery_info(entities=entities)
    await asyncio.sleep(0.1)

    # Remove the first entiry
    entities.pop(1)

    with patch("mqtt_entity.client._LOGGER") as mock_log:
        await mqc.publish_discovery_info(entities=entities)
        mock_log.info.assert_called_with(
            "Removing HASS MQTT discovery info %s",
            f"homeassistant/sensor/id_{root}/sen1/config",
        )

        mock_log.info.reset_mock()

        entities.pop(1)
        await mqc.remove_discovery_info(device_ids=[f"id_{root}"], keep_topics=[])
        assert mock_log.info.call_count == 2
        mock_log.info.assert_any_call(
            "Removing HASS MQTT discovery info %s",
            f"homeassistant/sensor/id_{root}/sen0/config",
        )
        mock_log.info.assert_any_call(
            "Removing HASS MQTT discovery info %s",
            f"homeassistant/sensor/id_{root}/sen2/config",
        )

    await mqc.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_connect(caplog):
    """Test connect."""
    with patch("mqtt_entity.client.Client") as client:
        mmock = MagicMock()
        mmock.is_connected.side_effect = [False, False, True]

        client.return_value = mmock
        mqc = mqtt_entity.MQTTClient()

        # assert not mmock.is_connected()
        # assert not cl._client.is_connected()
        # assert cl._client == cmock

        _LOGGER.error(dir(client))
        assert isinstance(mqc._client, Mock)

        await mqc.connect(None)

        assert 3 == mmock.is_connected.call_count

        mmock.is_connected.assert_called()

        assert "Connection" not in caplog.text
        _mqtt_on_connect(None, None, None, 1)
        # assert "Connection" in caplog.text



"""Utilities."""
from math import modf
from typing import Any


def required(_obj: Any, attr_obj: Any, val: Any) -> None:
    """An attrs property validator, mostly used in child classes."""
    if val is None:
        raise TypeError(f"Argument '{getattr(attr_obj, 'name', '')} missing'")


BOOL_ON = "ON"
BOOL_OFF = "OFF"


def tostr(val: Any) -> str:
    """Convert a value to a string with maximum 3 decimal places."""
    if val is None:
        return ""
    if isinstance(val, bool):
        return BOOL_ON if val else BOOL_OFF
    if not isinstance(val, float):
        return str(val)
    if modf(val)[0] == 0:
        return str(int(val))
    return f"{val:.3f}".rstrip("0")
